from aardvark import create_app, db
from aardvark.updater import update_account
from swag_client.swag import get_all_accounts
from tqdm import tqdm
from bunch import Bunch
from flask_script import Manager, Command, Option
from gunicorn.config import make_settings
import better_exceptions


manager = Manager(create_app)


@manager.command
def drop_db():
    """ Drops the database. """
    db.drop_all()


@manager.command
def create_db():
    """ Creates the database. """
    from aardvark.model import AWSIAMObject, AdvisorData
    db.create_all()


@manager.command
def config():
    """
    Creates a config file.

    SWAG_BUCKET = '...'
    ROLENAME = 'SecurityMonkey'
    SQLALCHEMY_DATABASE_URI = 'postgresql://user:pass@localhost:5432/db'
    SQLALCHEMY_DATABASE_URI = 'sqlite:///tmp/aardvark.db'
    PHANTOMJS = '/usr/local/bin/phantomjs'
    """
    bucket = raw_input('SWAG_BUCKET: ')
    role_name = raw_input('ROLENAME: ')
    db_uri = raw_input('DATABASE: ')
    phantom = raw_input('Path to phantomjs: ')

    log = """LOG_CFG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s %(levelname)s: %(message)s '
                '[in %(pathname)s:%(lineno)d]'
        }
    },
    'handlers': {
        'file': {
            'class': 'logging.handlers.RotatingFileHandler',
            'level': 'DEBUG',
            'formatter': 'standard',
            'filename': 'aardvark.log',
            'maxBytes': 10485760,
            'backupCount': 100,
            'encoding': 'utf8'
        },
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'standard',
            'stream': 'ext://sys.stdout'
        }
    },
    'loggers': {
        'aardvark': {
            'handlers': ['file', 'console'],
            'level': 'DEBUG'
        }
    }
}"""

    with open('config.py', 'w') as filedata:
        filedata.write('# Autogenerated config file\n')
        filedata.write('SWAG_BUCKET = "{bucket}"\n'.format(bucket=bucket))
        filedata.write('ROLENAME = "{role}"\n'.format(role=role_name))
        filedata.write('SQLALCHEMY_DATABASE_URI = "{uri}"\n'.format(uri=db_uri))
        filedata.write('PHANTOMJS = "{phantom}"\n'.format(phantom=phantom))
        filedata.write('SQLALCHEMY_TRACK_MODIFICATIONS = False\n')
        filedata.write(log)


@manager.option('-a', '--accounts', dest='accounts', type=unicode, default='all')
@manager.option('-r', '--arns', dest='arns', type=unicode, default='all')
def update(accounts, arns):
    """
    Asks AWS for new Access Advisor information.
    """
    accounts = _prep_accounts(accounts)
    arns = arns.split(',')
    app = create_app()
    with app.app_context():
        for account in tqdm(accounts, desc="Updating Accounts"):
            update_account(account.metadata['account_number'], app.config.get('ROLENAME'), arns)


def _prep_accounts(account_names):
    """
    Convert CLI provided account names into list of accounts from SWAG.
    Considers account aliases as well as account names.
    Returns a set of Bunch'd accounts.
    """
    app = create_app()
    with app.app_context():
        print('getting bucket {}'.format(app.config.get('SWAG_BUCKET')))
        service = {'services': {'security_monkey': {'enabled': True}}, "ours": True}
        accounts = get_all_accounts(bucket=app.config.get('SWAG_BUCKET'), **service).get('accounts')
    if account_names == 'all':
        return [Bunch(account) for account in accounts]

    lookup = {account['name']: Bunch(account) for account in accounts}
    for account in accounts:
        for alias in account['alias']:
            lookup[alias] = Bunch(account)

    matching_accounts = list()
    account_names = account_names.split(',')
    account_names = {name.lower() for name in account_names}
    for name in account_names:
        if name not in lookup:
            print('Could not find an account named %s' % name)
            continue

        matching_accounts.append(lookup[name])

    return matching_accounts


class GunicornServer(Command):
    """
    This is the main GunicornServer server, it runs the flask app with gunicorn and
    uses any configuration options passed to it.
    You can pass all standard gunicorn flags to this command as if you were
    running gunicorn itself.
    For example:
    aardvark start_api -w 4 -b 127.0.0.0:8002
    Will start gunicorn with 4 workers bound to 127.0.0.0:8002
    """
    description = 'Run the app within Gunicorn'

    def get_options(self):
        settings = make_settings()
        options = []
        for setting, klass in settings.items():
            if klass.cli:
                if klass.action:
                    if klass.action == 'store_const':
                        options.append(Option(*klass.cli, const=klass.const, action=klass.action))
                    else:
                        options.append(Option(*klass.cli, action=klass.action))
                else:
                    options.append(Option(*klass.cli))

        return options

    def run(self, *args, **kwargs):
        from gunicorn.app.wsgiapp import WSGIApplication
        from flask import current_app

        app = WSGIApplication()

        app.app_uri = 'aardvark:create_app()'
        return app.run()


def main():
    manager.add_command("start_api", GunicornServer())
    manager.run()


if __name__ == '__main__':
    manager.add_command("start_api", GunicornServer())
    manager.run()



__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "aardvark"
__summary__ = ("Multi-Account AWS IAM Access Advisor API")
__uri__ = "https://github.com/Netflix-Skunkworks/aardvark"

__version__ = "0.0.1dev2"

__author__ = "Patrick Kelley"
__email__ = "pkelley@netflix.com"

__license__ = "Apache License, Version 2.0"
__copyright__ = "Copyright 2017 {0}".format(__author__)


from flask import abort, jsonify
from aardvark.model import AWSIAMObject
from flask import Blueprint
from flask_restful import Api, Resource, reqparse
import better_exceptions  # noqa
import re


mod = Blueprint('advisor', __name__)
api = Api(mod)


class RoleGet(Resource):
    """
    Get item by id
    """
    def get(self, item_id):
        item = AWSIAMObject.query.filter(AWSIAMObject.id == item_id).scalar()
        if not item:
            abort(404)

        values = []
        for advisor_data in item.usage:
            values.append(dict(
                lastAuthenticated=advisor_data.lastAuthenticated,
                serviceName=advisor_data.serviceName,
                serviceNamespace=advisor_data.serviceNamespace
            ))

        return jsonify({item.arn: values})


class RoleSearch(Resource):
    """
    Search for roles by phrase, regex, or by ARN.
    """
    def __init__(self):
        super(RoleSearch, self).__init__()
        self.reqparse = reqparse.RequestParser()

    def get(self):
        self.reqparse.add_argument('page', type=int, default=1, location='args')
        self.reqparse.add_argument('count', type=int, default=30, location='args')
        self.reqparse.add_argument('phrase', type=str, default=None, location='args')
        self.reqparse.add_argument('regex', type=str, default=None, location='args')
        self.reqparse.add_argument('arn', type=str, default=None, location='args', action='append')
        args = self.reqparse.parse_args()
        page = args.pop('page')
        count = args.pop('count')
        phrase = args.pop('phrase', '')
        arns = args.pop('arn', [])
        regex = args.pop('regex', '')

        items = None

        if phrase:
            query = AWSIAMObject.query.filter(AWSIAMObject.arn.ilike('%'+phrase+'%'))

        if arns:
            query = AWSIAMObject.query.filter(AWSIAMObject.arn.in_(arns))

        # if regex:
        #     regex = re.compile(regex)
        #     all_items = AWSIAMObject.query.all()
        #     for item in all_items:
        #         if regex.match(item.arn):
        #             items.append(item)
        if regex:
            regex = re.compile(regex)
            query = AWSIAMObject.query.filter(AWSIAMObject.arn.regexp(regex))

        items = query.paginate(page, count)

        if not items:
            items = AWSIAMObject.query.paginate(page, count)

        values = dict(page=items.page, total=items.total, count=len(items.items))
        for item in items.items:
            item_values = []
            for advisor_data in item.usage:
                item_values.append(dict(
                    lastAuthenticated=advisor_data.lastAuthenticated,
                    serviceName=advisor_data.serviceName,
                    serviceNamespace=advisor_data.serviceNamespace
                ))
            values[item.arn] = item_values

        return jsonify(values)


api.add_resource(RoleGet, '/advisor/<int:item_id>')
api.add_resource(RoleSearch, '/advisors')


from aardvark import create_app
from cloudaux.aws.iam import list_roles
from cloudaux.aws.sts import boto3_cached_conn
import requests
import json
import os
import urllib
import subprocess

federation_base_url = 'https://signin.aws.amazon.com/federation'


def update_account(account_number, role_name, arns):
    """
    Updates Access Advisor data for a given AWS account.
    
    1) Gets list of IAM Role ARNs in target account.
    2) Gets IAM credentials in target account.
    3) Exchanges IAM credentials for Signin Token.
    4) Calls PhantomJS to do the dirty work.
    5) Saves PhantomJS output to our DB.
    
    :return: None
    """
    conn_details = {
        'account_number': account_number,
        'assume_role': role_name,
        'session_name': 'aardvark',
        'region': 'us-east-1'
    }
    arns = get_arns(arns, conn_details)
    if not arns:
        app = create_app()
        with app.app_context():
            app.logger.warn("Zero ARNs collected. Exiting")
        exit(-1)

    creds = get_creds(account_number, role_name)
    token = get_signin_token(creds)
    output_file = 'iam_access_advisor_{acct}'.format(acct=account_number)
    if call_phantom(token, list(arns), output_file) == 0:
        persist(output_file)


def get_arns(arns, conn):
    """
    Gets a list of all Role ARNs in a given account, optionally limited by a list of ARNS passed in.
    
    :return: list of role ARNs
    """
    roles = list_roles(**conn)
    account_arns = set([role['Arn'] for role in roles])
    result_arns = set()
    for arn in arns:
        if arn.lower() == 'all':
            return account_arns

        if arn not in account_arns:
            app = create_app()
            with app.app_context():
                app.logger.warn("Provided ARN {arn} not found in account.".format(arn=arn))
            continue

        result_arns.add(arn)

    return list(result_arns)


def get_creds(account_number, role_name):
    """
    Assumes into the target account and obtains Access Key, Secret Key, and Token
    
    :return: URL-encoded dictionary containing Access Key, Secret Key, and Token
    """
    client, credentials = boto3_cached_conn(
        'iam', account_number=account_number, assume_role=role_name, return_credentials=True)

    creds = json.dumps(dict(
        sessionId=credentials['AccessKeyId'],
        sessionKey=credentials['SecretAccessKey'],
        sessionToken=credentials['SessionToken']
    ))
    creds = urllib.quote(creds, safe='')
    return creds


def get_signin_token(creds):
    """
    Exchanges credentials dictionary for a signin token.
    
    1) Creates URL using credentials dictionary.
    2) Sends a GET request to that URL and parses the response looking for
    a signin token.
    
    :return: Signin Token
    """
    url = '{base}?Action=getSigninToken&Session={creds}'
    url = url.format(base=federation_base_url, creds=creds)
    return requests.get(url).json()['SigninToken']


def call_phantom(token, arns, output_file):
    """
    shells out to phantomjs.
    - Writes ARNs to a file that phantomjs will read as an input.
    - Phantomjs exchanges the token for session cookies.
    - Phantomjs then navigates to the IAM page and executes JavaScript
    to call GenerateServiceLastAccessedDetails for each ARN.
    - Every 10 seconds, Phantomjs calls GetServiceLastAccessedDetails
    - Phantom saves output to a file that is used by `persist()`

    :return: Exit code from phantomjs subprocess
    """

    path = os.path.dirname(__file__)
    console_js = os.path.join(path, 'awsconsole.js')
    arns_file = os.path.join(path, '_arns.json')
    with open(arns_file, 'w') as jsondata:
        json.dump(arns, jsondata)
    app = create_app()
    with app.app_context():
        return subprocess.call(
            [
                app.config.get('PHANTOMJS'),
                console_js,
                token,
                arns_file,
                output_file
            ],
            shell=False
        )


def persist(output_file):
    """
    Reads access advisor JSON file & persists to our database
    """
    from aardvark.model import AWSIAMObject, AdvisorData

    with open(output_file) as file_data:
        aa = json.load(file_data)

    for arn, data in aa.items():
        item = AWSIAMObject.get_or_create(arn)
        for service in data:
            AdvisorData.create_or_update(item.id,
                                         service['lastAuthenticated'],
                                         service['serviceName'],
                                         service['serviceNamespace'])




from flask_sqlalchemy import SQLAlchemy
from flask import Flask
import os.path

db = SQLAlchemy()

from aardvark.view import mod as advisor_bp  # noqa

BLUEPRINTS = [
    advisor_bp
]

API_VERSION = '1'


def create_app(config=None):
    app = Flask(__name__, static_url_path='/static')

    if config:
        app.config.from_pyfile(config)
    else:
        path = os.path.join(os.getcwd(), 'config.py')
        if os.path.isfile(path):
            app.config.from_pyfile(path)
        else:
            print('No config')
            app.config.from_pyfile('_config.py')

    # For ELB and/or Eureka
    @app.route('/healthcheck')
    def healthcheck():
        return 'ok'

    # Blueprints
    for bp in BLUEPRINTS:
        app.register_blueprint(bp, url_prefix="/api/{0}".format(API_VERSION))

    # Extensions:
    db.init_app(app)

    return app



from aardvark import db
from sqlalchemy import Column, Integer, String, Text
from sqlalchemy.orm import relationship
from sqlalchemy.schema import ForeignKey
# DateTime, Boolean, Unicode, Text


class AWSIAMObject(db.Model):
    """
    Meant to model AWS IAM Object Access Advisor.
    """
    __tablename__ = "aws_iam_object"
    id = Column(Integer, primary_key=True)
    arn = Column(Text(), nullable=True, index=True, unique=True)
    usage = relationship("AdvisorData", backref="item", cascade="all, delete, delete-orphan", foreign_keys="AdvisorData.item_id")

    @staticmethod
    def get_or_create(arn):
        item = AWSIAMObject.query.filter(AWSIAMObject.arn==arn).scalar()
        if not item:
            item = AWSIAMObject(arn=arn)
            db.session.add(item)
            db.session.commit()
            db.session.refresh(item)
        return item


class AdvisorData(db.Model):
    """
    Models certain IAM Access Advisor Data fields.

    {
      "totalAuthenticatedEntities": 1,
      "lastAuthenticatedEntity": "arn:aws:iam::XXXXXXXX:role/name",
      "serviceName": "Amazon Simple Systems Manager",
      "lastAuthenticated": 1489176000000,
      "serviceNamespace": "ssm"
    }
    """
    __tablename__ = "advisor_data"
    id = Column(Integer, primary_key=True)
    item_id = Column(Integer, ForeignKey("aws_iam_object.id"), nullable=False, index=True)
    lastAuthenticated = Column(Integer)
    serviceName = Column(String(128), index=True)
    serviceNamespace = Column(String(64), index=True)

    @staticmethod
    def create_or_update(item_id, lastAuthenticated, serviceNae, serviceNamespace):
        item = AdvisorData.query.filter(AdvisorData.item_id==item_id) \
            .filter(AdvisorData.serviceNamespace==serviceNamespace).scalar()

        if not item:
            item = AdvisorData(item_id=item_id, 
                               lastAuthenticated=lastAuthenticated,
                               serviceName=serviceNae,
                               serviceNamespace=serviceNamespace)
            db.session.add(item)
            db.session.commit()
            return

        if lastAuthenticated > item.lastAuthenticated:
            item.lastAuthenticated = lastAuthenticated
            db.session.add(item)
            db.session.commit()



"""
Aardvark
=====
Multi-Account AWS IAM Access Advisor API
:copyright: (c) 2017 by Netflix
:license: Apache, see LICENSE for more details.
"""
from __future__ import absolute_import

import sys
import os.path

from setuptools import setup, find_packages


ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))

# When executing the setup.py, we need to be able to import ourselves, this
# means that we need to add the src/ directory to the sys.path.
sys.path.insert(0, ROOT)

about = {}
with open(os.path.join(ROOT, "aardvark", "__about__.py")) as f:
    exec(f.read(), about)


install_requires = [
    'Flask-SQLAlchemy',
    'Bunch',
    'tqdm',
    'swag-client',
    'cloudaux>=1.1.7',
    'Flask-Script',
    'Flask',
    'Flask-RESTful',
    'gunicorn',
    'better_exceptions'
]

tests_require = [
]

docs_require = [
]

dev_requires = [
]


setup(
    name=about["__title__"],
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__email__"],
    url=about["__uri__"],
    description=about["__summary__"],
    long_description=open(os.path.join(ROOT, 'README.md')).read(),
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
        'tests': tests_require,
        'docs': docs_require,
        'dev': dev_requires,
    },
    entry_points={
        'console_scripts': [
            'aardvark = aardvark.manage:main',
        ],
    }
)